import itertools

import numpy as np
import pandas as pd
import scipy.stats


def _power(X, power):
    return X ** power if isinstance(X, np.ndarray) else X.power(power)


def asarray(x):
    return x.toarray() if scipy.sparse.issparse(x) else x


class DE:
    def __init__(
        self,
        series: pd.Series,
        nfeatures: int,
        batch_size: int,
        get_batch_fn,
        base: float = None,
        one_vs_rest: bool = True,
    ):
        """

        :param series: Categorical series in adata.obs to group by
        :param nfeatures: Number of features in adata
        :param batch_size: Number of features per batch
        :param get_batch_fn: Function to retrieve data from a batch
        :param base: adata.uns['log1p']['base']
        :param one_vs_rest: Whether to compare each group vs rest or all pairs of groups
        """
        mean_df = None
        variance_df = None
        frac_expressed_df = None
        indicator_df = pd.get_dummies(series)
        if one_vs_rest:
            pairs = []
            rest_indicator_df = pd.DataFrame()
            for c in indicator_df:
                rest_name = str(c) + "_rest"
                if rest_name in indicator_df:
                    counter = 1
                    rest_name = str(c) + "_rest-{}".format(counter)
                    while rest_name in indicator_df:
                        counter = counter + 1
                        rest_name = str(c) + "_rest-{}".format(counter)
                pairs.append((c, rest_name))
                rest_indicator_series = indicator_df[c].astype(bool)
                rest_indicator_series = ~rest_indicator_series
                rest_indicator_df[rest_name] = rest_indicator_series.astype(int)
            indicator_df = indicator_df.join(rest_indicator_df)
        else:
            pairs = list(itertools.combinations(series.cat.categories, 2))
        count_ = indicator_df.sum(axis=0)  # count per group
        count_values = count_.values
        A = scipy.sparse.coo_matrix(indicator_df.astype(float).T)
        n_row = A.shape[0]
        row_sums = np.asarray(A.sum(axis=1))
        D = scipy.sparse.dia_matrix(((row_sums.T**-1), [0]), shape=(n_row, n_row))
        A = D * A
        dof = 1

        for i in range(0, nfeatures, batch_size):
            adata_batch = get_batch_fn(i)
            X = adata_batch.X
            mean_ = asarray(A @ X)  # (groups, genes)
            mean_sq = asarray(A @ _power(X, 2))
            sq_mean = mean_**2
            var_ = mean_sq - sq_mean
            # enforce R convention (unbiased estimator) for variance
            precision = 2 << (42 if X.dtype == np.float64 else 20)
            # detects loss of precision in mean_sq - sq_mean, which suggests variance is 0
            var_[precision * var_ < sq_mean] = 0
            if dof != 0:
                var_ *= (count_values / (count_values - dof))[:, np.newaxis]

            frac_expressed_ = None
            if scipy.sparse.issparse(X):
                frac_expressed_ = asarray(A @ (X != 0))
            _mean_df = pd.DataFrame(
                mean_, columns=adata_batch.var.index, index=indicator_df.columns
            )
            _variance_df = pd.DataFrame(
                var_, columns=adata_batch.var.index, index=indicator_df.columns
            )

            # groups on rows, genes on columns
            mean_df = pd.concat((mean_df, _mean_df), axis=1) if mean_df is not None else _mean_df
            variance_df = (
                pd.concat((variance_df, _variance_df), axis=1)
                if variance_df is not None
                else _variance_df
            )
            if frac_expressed_ is not None:
                _frac_expressed_df = pd.DataFrame(
                    frac_expressed_, columns=adata_batch.var.index, index=indicator_df.columns
                )
                frac_expressed_df = (
                    pd.concat((frac_expressed_df, _frac_expressed_df), axis=1)
                    if frac_expressed_df is not None
                    else _frac_expressed_df
                )

        if base is not None:
            expm1_func = lambda x: np.expm1(x * np.log(base))
        else:
            expm1_func = np.expm1
        pair2results = dict()
        for p in pairs:
            group_one, group_two = p
            nobs1 = count_.loc[group_one]
            nobs2 = count_.loc[group_two]

            # add small value to remove 0's
            foldchanges = np.log2(
                (expm1_func(mean_df.loc[group_one].values) + 1e-9)
                / (expm1_func(mean_df.loc[group_two].values) + 1e-9)
            )
            with np.errstate(invalid="ignore"):
                scores, pvals = scipy.stats.ttest_ind_from_stats(
                    mean1=mean_df.loc[group_one],
                    std1=np.sqrt(variance_df.loc[group_one]),
                    nobs1=nobs1,
                    mean2=mean_df.loc[group_two],
                    std2=np.sqrt(variance_df.loc[group_two]),
                    nobs2=nobs2,
                    equal_var=False,  # Welch's
                )

            scores[np.isnan(scores)] = 0
            pvals[np.isnan(pvals)] = 1
            key = p[0] if one_vs_rest else p
            pair2results[key] = dict(
                scores=scores,
                pvals=pvals,
                logfoldchanges=foldchanges,
                frac_expressed1=(
                    frac_expressed_df.loc[group_one].values
                    if frac_expressed_df is not None
                    else None
                ),
                frac_expressed2=(
                    frac_expressed_df.loc[group_two].values
                    if frac_expressed_df is not None
                    else None
                ),
            )
        self.pair2results = pair2results



import os
from urllib.parse import urlparse

import numpy as np
import fsspec
import pandas as pd
import pandas._libs.json as ujson
from flask import make_response

from cirrocumulus.envir import CIRRO_DATASET_PROVIDERS


try:
    dumps = ujson.dumps
except AttributeError:
    dumps = ujson.ujson_dumps


def add_dataset_providers():
    from cirrocumulus.api import dataset_api

    dataset_providers = []

    for p in os.environ[CIRRO_DATASET_PROVIDERS].split(","):
        try:
            dataset_api.add(create_instance(p))
            dataset_providers.append(p)
        except ModuleNotFoundError:  # ignore if required libraries are not installed
            pass

    os.environ[CIRRO_DATASET_PROVIDERS] = ",".join(dataset_providers)


def import_path(name):
    import importlib

    dot_index = name.rfind(".")
    return getattr(importlib.import_module(name[0:dot_index]), name[dot_index + 1 :])


def create_instance(class_name):
    return import_path(class_name)()


def get_scheme(path):
    pr = urlparse(path)
    if len(pr.scheme) <= 1:  # for file paths: /foo/bar/test.h5ad or C:/foo/bar/test.h5ad
        return "file"
    return pr.scheme


fsspec_kwargs = dict(s3_additional_kwargs=dict(ACL="bucket-owner-full-control"))


def get_fs(path):
    return fsspec.filesystem(get_scheme(path), **fsspec_kwargs)


def open_file(urlpath, mode="rb", compression=None):
    return fsspec.open(urlpath, mode=mode, compression=compression, **fsspec_kwargs)


def to_json(data, orient="values"):
    return dumps(data, double_precision=2, orient=orient)


def json_response(data, response=200):
    # response = make_response(simplejson.dumps(data, check_circular=True), response)
    # response = make_response(json.dumps(data), response)
    s = dumps(data, double_precision=2, orient="values")
    # s = nujson.dumps(data, double_precision=1)
    response = make_response(s, response)
    response.headers["Content-Type"] = "application/json"
    return response


def get_email_domain(email):
    at_index = email.find("@")
    domain = None
    if at_index != -1:
        domain = email[at_index + 1 :]
    return domain


def load_dataset_schema(url):
    import os
    import json

    def get_extension(path):
        name, ext = os.path.splitext(path)
        if ext == ".gz":
            name, ext = os.path.splitext(name)
            if ext == ".json":
                ext = ".json.gz"
        return ext

    extension = get_extension(url)
    json_schema = None
    if extension in [".json", ".json.gz", ""]:
        scheme = get_scheme(url)
        fs = get_fs(scheme)
        if extension == "":
            url = os.path.join(url, "index.json.gz")
            extension = get_extension(url)
        if extension == ".json.gz":
            import gzip

            with gzip.open(fs.open(url)) as f:
                json_schema = json.load(f)
        else:
            with fs.open(url) as f:
                json_schema = json.load(f)
    return json_schema


def write_top_half_gct(f, row_metadata_df, col_metadata_df, metadata_null, filler_null):
    """Write the top half of the gct file: top-left filler values, row metadata headers, and top-
    right column metadata.

    Args:
        f (file handle): handle for output file
        row_metadata_df (pandas df)
        col_metadata_df (pandas df)
        metadata_null (string): how to represent missing values in the metadata
        filler_null (string): what value to fill the top-left filler block with
    Returns:
        None
    """
    # Initialize the top half of the gct including the third line
    size_of_top_half_df = (
        1 + col_metadata_df.shape[1],
        1 + row_metadata_df.shape[1] + col_metadata_df.shape[0],
    )

    top_half_df = pd.DataFrame(np.full(size_of_top_half_df, filler_null, dtype=object))

    # Assemble the third line of the gct: "id", then rhds, then cids
    top_half_df.iloc[0, :] = np.hstack(
        ("id", row_metadata_df.columns.values, col_metadata_df.index.values)
    )

    # Insert the chds
    top_half_df.iloc[range(1, top_half_df.shape[0]), 0] = col_metadata_df.columns.values

    # Insert the column metadata, but first convert to strings and replace NaNs
    col_metadata_indices = (
        range(1, top_half_df.shape[0]),
        range(1 + row_metadata_df.shape[1], top_half_df.shape[1]),
    )
    # pd.DataFrame.at to insert into dataframe(python3)
    top_half_df.at[col_metadata_indices[0], col_metadata_indices[1]] = (
        col_metadata_df.astype(str).replace("nan", value=metadata_null).T.values
    )

    # Write top_half_df to file
    top_half_df.to_csv(f, header=False, index=False, sep="\t")


def write_bottom_half_gct(f, row_metadata_df, data_df, data_null, data_float_format, metadata_null):
    """Write the bottom half of the gct file: row metadata and data.

    Args:
        f (file handle): handle for output file
        row_metadata_df (pandas df)
        data_df (pandas df)
        data_null (string): how to represent missing values in the data
        metadata_null (string): how to represent missing values in the metadata
        data_float_format (string): how many decimal points to keep in representing data
    Returns:
        None
    """
    # create the left side of the bottom half of the gct (for the row metadata)
    size_of_left_bottom_half_df = (row_metadata_df.shape[0], 1 + row_metadata_df.shape[1])
    left_bottom_half_df = pd.DataFrame(
        np.full(size_of_left_bottom_half_df, metadata_null, dtype=object)
    )

    # create the full bottom half by combining with the above with the matrix data
    bottom_half_df = pd.concat([left_bottom_half_df, data_df.reset_index(drop=True)], axis=1)
    bottom_half_df.columns = range(bottom_half_df.shape[1])

    # Insert the rids
    bottom_half_df.iloc[:, 0] = row_metadata_df.index.values

    # Insert the row metadata, but first convert to strings and replace NaNs
    row_metadata_col_indices = range(1, 1 + row_metadata_df.shape[1])
    bottom_half_df.iloc[:, row_metadata_col_indices] = (
        row_metadata_df.astype(str).replace("nan", value=metadata_null).values
    )

    # Write bottom_half_df to file
    bottom_half_df.to_csv(
        f, header=False, index=False, sep="\t", na_rep=data_null, float_format=data_float_format
    )


def adata2gct(adata, f):
    data_float_format = "%.4f"
    f.write("#1.3\n")
    f.write(
        str(adata.shape[0])
        + "\t"
        + str(adata.shape[1])
        + "\t"
        + str(adata.obs.shape[1])
        + "\t"
        + str(adata.var.shape[1])
        + "\n"
    )
    write_top_half_gct(f, adata.obs, adata.var, "", "")
    write_bottom_half_gct(f, adata.obs, pd.DataFrame(adata.X), "", data_float_format, "")



import os
import math
import logging

import pandas as pd

from cirrocumulus.diff_exp import DE
from cirrocumulus.ot.transport_map_model import read_transport_map_dir
from cirrocumulus.util import dumps

from .data_processing import get_filter_str, get_mask, get_selected_data
from .envir import (
    CIRRO_DATABASE_CLASS,
    CIRRO_JOB_RESULTS,
    CIRRO_JOB_TYPE,
    CIRRO_MAX_WORKERS,
    CIRRO_SERVE,
)
from .fdr import fdrcorrection
from .util import add_dataset_providers, create_instance, get_fs, import_path, open_file


executor = None
job_id_2_future = dict()

logger = logging.getLogger("cirro")


def save_job_result_to_file(result, job_id):
    new_result = dict()
    new_result["content-type"] = result.pop("content-type")
    if new_result["content-type"] == "application/json":
        new_result["content-encoding"] = "gzip"
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + ".json.gz")
        with open_file(url, "wt", compression="gzip") as out:
            out.write(dumps(result, double_precision=2, orient="values"))
    elif new_result["content-type"] == "application/h5ad":
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + ".h5ad")
        with get_fs(url).open(url, "wb") as out:
            result["data"].write(out)
    elif new_result["content-type"] == "application/zarr":
        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + ".zarr")
        result["data"].write_zarr(get_fs(url).get_mapper(url))
    elif new_result["content-type"] == "application/parquet":
        import pyarrow as pa
        import pyarrow.parquet as pq

        url = os.path.join(os.environ[CIRRO_JOB_RESULTS], str(job_id) + ".parquet")
        pq.write_table(pa.Table.from_pandas(result["data"]), url, filesystem=get_fs(url))
    else:
        raise ValueError("Unknown content-type {}".format(new_result["content-type"]))
    new_result["url"] = url
    return new_result


def delete_job(job_id):
    future = job_id_2_future.get(job_id)
    if future is not None and not future.done():
        del job_id_2_future[job_id]
        future.cancel()
        logger.info("Cancel job {}".format(job_id))


def done_callback(future):
    for job_id in list(job_id_2_future.keys()):
        if job_id_2_future[job_id] == future:
            del job_id_2_future[job_id]
            logger.info("Job {} done".format(job_id))
            break


def submit_job(database_api, dataset_api, email, dataset, job_name, job_type, params):
    global executor

    is_serve = os.environ.get(CIRRO_SERVE) == "true"
    if executor is None:
        max_workers = int(os.environ.get(CIRRO_MAX_WORKERS, "2" if is_serve else "1"))
        if max_workers > 0:
            import multiprocessing
            from concurrent.futures.process import ProcessPoolExecutor
            from concurrent.futures.thread import ThreadPoolExecutor

            executor = (
                ProcessPoolExecutor(
                    max_workers=max_workers, mp_context=multiprocessing.get_context("spawn")
                )
                if is_serve
                else ThreadPoolExecutor(max_workers=max_workers)
            )
    job_id = database_api.create_job(
        email=email, dataset_id=dataset["id"], job_name=job_name, job_type=job_type, params=params
    )
    if executor is not None:
        future = executor.submit(
            run_job,
            email,
            job_id,
            job_name,
            job_type,
            dataset,
            params,
            database_api if not is_serve else None,
            dataset_api if not is_serve else None,
        )
        future.add_done_callback(done_callback)
        job_id_2_future[job_id] = future
    else:
        run_job(
            email,
            job_id,
            job_name,
            job_type,
            dataset,
            params,
            database_api if not is_serve else None,
            dataset_api if not is_serve else None,
        )
    return job_id


def get_obs(dataset_api, dataset, dataset_info, params):
    obs_fields = params.get("obs")
    if obs_fields is not None:
        obs = dataset_api.read_dataset(keys=dict(obs=obs_fields), dataset=dataset).obs
        obs_field = obs_fields[0]
        if len(obs_fields) > 1:
            # combine in to one field
            obs_field = "_".join(obs_fields)
            obs[obs_field] = obs[obs_fields[0]].astype(str)
            for i in range(1, len(obs_fields)):
                obs[obs_field] += "_" + obs[obs_fields[i]].astype(str)
            obs[obs_field] = obs[obs_field].astype("category")
        return obs, obs_field
    else:
        filters = [params["filter"], params["filter2"]]
        filter_names = [get_filter_str(params["filter"]), get_filter_str(params["filter2"])]
        for i in range(len(filter_names)):
            if filter_names[i] is None:
                filter_names[i] = "group_" + str(i + 1)
        obs = pd.DataFrame(index=pd.RangeIndex(dataset_info["shape"][0]).astype(str))
        obs_field = "selection"  # order of categories needs to match filter names
        # obs[obs_field] = "3"
        masks, _ = get_mask(dataset_api, dataset, dataset_info, filters)
        for i in range(2):
            obs.loc[masks[i], obs_field] = filter_names[i]
        obs[obs_field] = (
            obs[obs_field].astype("category").cat.as_ordered().cat.reorder_categories(filter_names)
        )
        return obs, obs_field


def run_job(email, job_id, job_name, job_type, dataset, params, database_api, dataset_api):
    if database_api is None:
        database_api = create_instance(os.environ[CIRRO_DATABASE_CLASS])
    if dataset_api is None:
        from cirrocumulus.api import dataset_api

        add_dataset_providers()
    database_api.update_job(email=email, job_id=job_id, status="running", result=None)
    f = os.environ[CIRRO_JOB_TYPE + job_type]
    if f is None:
        database_api.update_job(email=email, job_id=job_id, status="error", result=None)
        raise ValueError("No function to handle {}".format(job_type))
    import_path(f)(email, job_id, job_name, job_type, dataset, params, database_api, dataset_api)


def run_ot_trajectory(
    email, job_id, job_name, job_type, dataset, params, database_api, dataset_api
):
    dataset_info = dataset_api.get_dataset_info(dataset)
    timepoint_field = dataset_info.get("timepoint_field", "day")
    selected_adata = get_selected_data(
        dataset_api,
        dataset,
        measures=["obs/index", "obs/" + timepoint_field],
        data_filter=params["filter"],
        dataset_info=dataset_info,
    )

    day = selected_adata.obs[timepoint_field].unique()
    if len(day) > 1:
        database_api.update_job(
            email=email, job_id=job_id, status="error", result="More than one timepoint selected"
        )
        raise ValueError("More than one timepoint selected")
    tmap_name = params["tmap"]
    for tmap_item in dataset["ot"]["tmaps"]:
        if tmap_item["name"] == tmap_name:
            break

    tmap_model = read_transport_map_dir(tmap_item["path"])
    populations = tmap_model.population_from_ids(selected_adata.obs["index"].values, at_time=day[0])
    populations[0].name = job_name

    trajectory_dataset = tmap_model.trajectories(populations)

    # match order in dataset
    # trajectory_ds = trajectory_ds[adata.obs.index]

    result = dict(data=trajectory_dataset.to_df())
    result["content-type"] = "application/parquet"
    database_api.update_job(email=email, job_id=job_id, status="complete", result=result)


def run_de(email, job_id, job_name, job_type, dataset, params, database_api, dataset_api):
    dataset_info = dataset_api.get_dataset_info(dataset)
    var_names = dataset_info["var"]
    nfeatures = len(var_names)
    batch_size = (
        math.ceil(nfeatures / 5) if os.environ.get(CIRRO_SERVE) == "true" else nfeatures
    )  # TODO more intelligent batches

    def get_batch_fn(i):
        end = min(nfeatures, i + batch_size)
        adata = dataset_api.read_dataset(keys=dict(X=[slice(i, end)]), dataset=dataset)
        if batch_size != nfeatures:
            frac = end / nfeatures
            status = "running {:.0f}%".format(100 * frac) if frac < 1 else "saving results"
            logger.info(status)
            database_api.update_job(email=email, job_id=job_id, status=status, result=None)
        return adata

    compare_pairs = params.get("pairs") == "1"
    obs, obs_field = get_obs(dataset_api, dataset, dataset_info, params)

    de_results = DE(
        series=obs[obs_field],
        nfeatures=nfeatures,
        batch_size=batch_size,
        get_batch_fn=get_batch_fn,
        one_vs_rest=not compare_pairs,
    )  # TODO get base

    # group:field is object entry
    result_df = pd.DataFrame(data={"index": var_names})
    has_frac_expressed = False
    comparison_names = []
    for comparison in de_results.pair2results.keys():
        result = de_results.pair2results[comparison]
        comparison_name = comparison if isinstance(comparison, str) else "_".join(comparison)
        comparison_names.append(comparison_name)

        pvals = fdrcorrection(result["pvals"])
        result_df[f"{comparison_name}:pvals_adj"] = pvals
        result_df[f"{comparison_name}:scores"] = result["scores"]
        result_df[f"{comparison_name}:lfc"] = result["logfoldchanges"]

        if result.get("frac_expressed1") is not None:
            has_frac_expressed = True
            result_df[f"{comparison_name}:pts_1"] = result["frac_expressed1"]
            result_df[f"{comparison_name}:pts_2"] = result["frac_expressed2"]
    # client expects field {comparison_name}:pvals_adj
    result = dict(
        groups=comparison_names,
        fields=["pvals_adj", "scores", "lfc"] + (["pts_1", "pts_2"] if has_frac_expressed else []),
        data=result_df.to_dict(orient="records"),
    )
    result["content-type"] = "application/json"
    database_api.update_job(email=email, job_id=job_id, status="complete", result=result)



import logging

import pandas as pd
import anndata
import scipy.sparse
from anndata import AnnData
from pandas import CategoricalDtype

from cirrocumulus.abstract_dataset import AbstractDataset
from cirrocumulus.anndata_util import ADATA_LAYERS_UNS_KEY, ADATA_MODULE_UNS_KEY, dataset_schema
from cirrocumulus.io_util import add_spatial, read_star_fusion_file


logger = logging.getLogger("cirro")

CATEGORICAL_FIELDS_CONVERT = ["seurat_clusters"]


def read_adata(path, filesystem, backed=False, spatial_directory=None, use_raw=False):
    path = path.rstrip(filesystem.sep)
    path_lc = path.lower()
    if path_lc.endswith(".loom"):
        adata = anndata.read_loom(filesystem.open(path))
    elif path_lc.endswith(".zarr"):
        adata = anndata.read_zarr(filesystem.get_mapper(path))
    elif path_lc.endswith(".tsv"):
        adata = read_star_fusion_file(filesystem.open(path))
    elif path_lc.endswith(".h5"):
        import scanpy as sc

        adata = sc.read_10x_h5(path)  # fsspec not supported
        sep_index = path.rfind(filesystem.sep)
        cells_path = "cells.parquet"
        analysis_path = "analysis"
        if sep_index != -1:
            cells_path = path[0:sep_index] + filesystem.sep + cells_path
            analysis_path = path[0:sep_index] + filesystem.sep + analysis_path

        if filesystem.exists(cells_path):  # xenium
            df = pd.read_parquet(cells_path).set_index("cell_id")
            df.index = df.index.astype(str)
            adata.obs = adata.obs.join(df)
            if "x_centroid" in adata.obs.columns:
                adata.obsm["X_spatial"] = adata.obs[["x_centroid", "y_centroid"]].values

                del adata.obs["x_centroid"]
                del adata.obs["y_centroid"]
        if filesystem.exists(analysis_path) and filesystem.isdir(analysis_path):  # xenium
            clustering_results = filesystem.glob(
                analysis_path + filesystem.sep + "clustering" + filesystem.sep + "**clusters.csv"
            )
            for clustering_result in clustering_results:
                df = pd.read_csv(clustering_result, index_col="Barcode")
                if len(df.columns) == 1:
                    df.index = df.index.astype(str)
                    df[df.columns[0]] = df[df.columns[0]].astype("category")
                    rename = dict()
                    # /analysis/clustering/gene_expression_graphclust/clusters.csv'
                    rename[df.columns[0]] = clustering_result.split(filesystem.sep)[-2]
                    df = df.rename(rename, axis=1)
                    adata.obs = adata.obs.join(df)
            umap_results = filesystem.glob(
                analysis_path + filesystem.sep + "umap" + filesystem.sep + "**projection.csv"
            )
            for umap_result in umap_results:
                df = pd.read_csv(umap_result, index_col="Barcode")
                if len(df.columns) == 2 or len(df.columns) == 3:
                    df.index = df.index.astype(str)
                    adata.obs = adata.obs.join(df)
                    adata.obsm["X_" + "_".join(umap_result.split(filesystem.sep)[-2:])] = adata.obs[
                        df.columns
                    ].values
                    for c in df.columns:
                        del adata.obs[c]

    elif path_lc.endswith(".rds"):  # Seurat, convert to h5ad
        h5_file = path + ".h5ad"
        import os

        if (
            not os.path.exists(h5_file)
            or abs(os.path.getmtime(h5_file) - os.path.getmtime(path)) > 0.00001
        ):
            import shutil
            import subprocess

            import pkg_resources

            print("Converting Seurat object")
            if os.path.exists(h5_file):
                os.remove(h5_file)
            subprocess.check_call(
                [
                    "Rscript",
                    pkg_resources.resource_filename("cirrocumulus", "seurat2h5ad.R"),
                    path,
                    h5_file,
                ]
            )
            shutil.copystat(path, h5_file)
        adata = anndata.read_h5ad(h5_file, backed="r" if backed else None)
        if adata.raw is not None and adata.shape[0] == adata.raw.shape[0]:
            print("Using adata.raw")
            adata = anndata.AnnData(
                X=adata.raw.X, var=adata.raw.var, obs=adata.obs, obsm=adata.obsm, uns=adata.uns
            )
    else:
        if backed:
            adata = anndata.read_h5ad(path, backed="r")
        else:
            adata = anndata.read_h5ad(filesystem.open(path))

    if "module" in adata.uns:
        adata.uns[ADATA_MODULE_UNS_KEY] = anndata.AnnData(
            X=adata.uns["module"]["X"], var=adata.uns["module"]["var"]
        )

    if "images" in adata.uns:
        images = adata.uns["images"]
        if isinstance(images, dict):
            adata.uns["images"] = [images]

    if use_raw and adata.raw is not None and adata.shape[0] == adata.raw.shape[0]:
        logger.info("Using adata.raw")
        adata = anndata.AnnData(
            X=adata.raw.X, var=adata.raw.var, obs=adata.obs, obsm=adata.obsm, uns=adata.uns
        )
    adata.var_names_make_unique()
    if spatial_directory is not None:
        if not add_spatial(adata, spatial_directory):
            logger.info("No spatial data found in {}".format(spatial_directory))

    for field in CATEGORICAL_FIELDS_CONVERT:
        if field in adata.obs and not isinstance(adata.obs[field].dtype, CategoricalDtype):
            logger.info("Converting {} to categorical".format(field))
            adata.obs[field] = adata.obs[field].astype(str).astype("category")
    return adata


class AnndataDataset(AbstractDataset):
    def __init__(self, backed=None):
        super().__init__()
        self.path_to_data = {}
        self.backed = backed

    def get_suffixes(self):
        return ["h5ad", "loom", "rds", "zarr", "h5"]

    def get_result(self, filesystem, path, dataset, result_id):
        adata = self.get_data(filesystem, path)
        if result_id in adata.uns:
            return str(adata.uns[result_id])
        return super().get_result(filesystem, path, dataset, result_id)

    def read_adata(self, filesystem, path):
        adata = read_adata(path, filesystem, self.backed)
        return adata

    def add_data(self, path, data):
        self.path_to_data[path] = data

    def get_data(self, filesystem, path):
        adata = self.path_to_data.get(path)
        if adata is None:
            adata = self.read_adata(filesystem, path)
            if scipy.sparse.isspmatrix_csr(adata.X) and adata.X.shape[1] > 1:
                adata.X = adata.X.tocsc()
            self.add_data(path, adata)
        return adata

    def get_schema(self, filesystem, path):
        adata = self.get_data(filesystem, path)
        schema = dataset_schema(adata)
        if "cirro-schema" in adata.uns:
            import json

            s = json.loads(str(adata.uns["cirro-schema"]))
            keys = ["markers", "results"]
            for key in keys:
                if key in s:
                    schema[key] = s[key]
        return schema

    @staticmethod
    def get_X(adata, keys, layer=None):
        if len(keys) == 1 and isinstance(keys[0], slice):  # special case if slice specified
            keys = keys[0]
        d = adata[:, keys]
        X = d.X if layer is None else d.layers[layer]
        if scipy.sparse.issparse(X) and not scipy.sparse.isspmatrix_csc(X):
            X = X.tocsc()

        var = pd.DataFrame(index=d.var.index)
        return X, var

    def read_dataset(self, filesystem, path, keys=None, dataset=None):
        adata = self.get_data(filesystem, path)
        if keys is None:
            keys = {}
        keys = keys.copy()
        X_keys = keys.pop("X", [])
        obs_keys = keys.pop("obs", [])
        basis_keys = keys.pop("basis", [])
        module_keys = keys.pop("module", [])
        X = None
        obs = None
        var = None
        obsm = {}
        adata_modules = None
        layers = {}
        for layer_key in keys.keys():
            X_layer, var_layer = AnndataDataset.get_X(adata, keys[layer_key], layer_key)
            adata_layer = AnnData(X=X_layer, var=var_layer)
            layers[layer_key] = adata_layer
        if len(X_keys) > 0:
            X, var = AnndataDataset.get_X(adata, X_keys)
        if len(module_keys) > 0:
            if len(module_keys) == 1 and isinstance(
                module_keys[0], slice
            ):  # special case if slice specified
                module_keys = module_keys[0]
            adata_modules = adata.uns[ADATA_MODULE_UNS_KEY][:, module_keys]

        if len(obs_keys) > 0:
            obs = pd.DataFrame()
            for key in obs_keys:
                if key == "index":
                    values = adata.obs.index.values
                else:
                    values = adata.obs[key].values
                obs[key] = values
        if len(basis_keys) > 0:
            for key in basis_keys:
                embedding_data = adata.obsm[key]
                obsm[key] = embedding_data
            if X is None:  # anndata requires empty X
                X = scipy.sparse.coo_matrix(([], ([], [])), shape=(embedding_data.shape[0], 0))
        if X is None and obs is None and len(obsm.keys()) == 0:
            obs = pd.DataFrame(index=pd.RangeIndex(adata.shape[0]).astype(str))

        adata = AnnData(X=X, obs=obs, var=var, obsm=obsm)
        if adata_modules is not None:
            adata.uns[ADATA_MODULE_UNS_KEY] = adata_modules
        adata.uns[ADATA_LAYERS_UNS_KEY] = layers
        return adata



CIRRO_TEST = "CIRRO_TEST"  # whether running using pytest
# authentication/authorization
CIRRO_AUTH_CLIENT_ID = "CIRRO_AUTH_CLIENT_ID"
CIRRO_AUTH_ISSUER = "CIRRO_AUTH_ISSUER"
CIRRO_AUTH = "CIRRO_AUTH"  # python class to validate tokens (created based on CIRRO_AUTH_PROVIDER)
CIRRO_AUTH_PROVIDER = "CIRRO_AUTH_PROVIDER"  # okta, google

CIRRO_DB_URI = "CIRRO_DB_URI"
CIRRO_EMAIL = "CIRRO_EMAIL"
CIRRO_SERVE = "CIRRO_SERVE"
CIRRO_MAX_WORKERS = "CIRRO_MAX_WORKERS"
CIRRO_FOOTER = "CIRRO_FOOTER"
CIRRO_BRAND = "CIRRO_BRAND"
CIRRO_UPLOAD = "CIRRO_UPLOAD"

CIRRO_COMPRESS = "CIRRO_COMPRESS"
CIRRO_JOB_RESULTS = "CIRRO_JOB_RESULTS"  # job result storage location
CIRRO_DATABASE_CLASS = "CIRRO_DATABASE_CLASS"
CIRRO_DATABASE = "CIRRO_DATABASE"
CIRRO_DATASET_PROVIDERS = "CIRRO_DATASET_PROVIDERS"

CIRRO_MIXPANEL = "CIRRO_MIXPANEL"
# for mounting a bucket locally. Comma separated string of bucket:local_path. Example s3://foo/bar:/fsx
CIRRO_MOUNT = "CIRRO_MOUNT"
CIRRO_LOG_LEVEL = "CIRRO_LOG_LEVEL"
# columns to display to user
CIRRO_DATASET_SELECTOR_COLUMNS = "CIRRO_DATASET_SELECTOR_COLUMNS"
CIRRO_JOB_TYPE = "CIRRO_JOB_TYPE"

# path to JSON file for library list when adding new dataset
CIRRO_LIBRARY = "CIRRO_LIBRARY"

# path to JSON file for species list when adding new dataset
CIRRO_SPECIES = "CIRRO_SPECIES"
# path to OBO file
CIRRO_CELL_ONTOLOGY = "CIRRO_CELL_ONTOLOGY"

# comma separated list of paths to allow all logged in users to download from
CIRRO_STATIC_DIR = "CIRRO_STATIC_DIR"

SERVER_CAPABILITY_RENAME_CATEGORIES = "SERVER_CAPABILITY_RENAME_CATEGORIES"
SERVER_CAPABILITY_JOBS = "SERVER_CAPABILITY_JOBS"
SERVER_CAPABILITY_FEATURE_SETS = "SERVER_CAPABILITY_FEATURE_SETS"
SERVER_CAPABILITY_LINKS = "SERVER_CAPABILITY_LINKS"
SERVER_CAPABILITY_EDIT_DATASET = "SERVER_CAPABILITY_EDIT_DATASET"
SERVER_CAPABILITY_ADD_DATASET = "SERVER_CAPABILITY_ADD_DATASET"
SERVER_CAPABILITY_DELETE_DATASET = "SERVER_CAPABILITY_DELETE_DATASET"



import os

from cirrocumulus.util import get_fs


def get_path(dataset, dataset_path):
    path = dataset["url"]
    if path[len(path) - 1] == "/":  # remove trailing slash
        path = path[0 : len(path) - 1]
    if path.endswith(".json"):
        path = os.path.dirname(path)
    path = path + "/" + dataset_path
    return path


class DatasetAPI:
    def __init__(self):
        self.suffix_to_provider = {}
        self.default_provider = None
        self.cached_dataset_info = None
        self.cached_dataset_id = None

    def get_dataset_provider(self, path):
        index = path.rfind(".")
        if index == -1:
            return self.default_provider
        suffix = path[index + 1 :].lower()
        suffix = suffix.rstrip("/")
        provider = self.suffix_to_provider.get(suffix)
        return provider if provider is not None else self.default_provider

    def add(self, provider):
        if self.default_provider is None:
            self.default_provider = provider
        suffixes = provider.get_suffixes()
        for suffix in suffixes:
            self.suffix_to_provider[suffix.lower()] = provider

    def get_dataset_info(self, dataset):
        dataset_id = dataset["id"]
        if self.cached_dataset_id == dataset_id:
            dataset_info = self.cached_dataset_info
        else:
            path = dataset["url"]
            provider = self.get_dataset_provider(path)
            dataset_info = provider.get_dataset_info(get_fs(path), path)
            self.cached_dataset_info = dataset_info
            self.cached_dataset_id = dataset_id
        return dataset_info

    def get_schema(self, dataset):
        path = dataset["url"]
        provider = self.get_dataset_provider(path)
        schema_dict = provider.get_schema(get_fs(path), path)
        if "summary" in dataset:
            schema_dict["summary"] = dataset["summary"]
        if "markers" in schema_dict:
            schema_dict["markers_read_only"] = schema_dict.pop("markers")
        return schema_dict

    def read_dataset(self, dataset, keys=[]):
        path = dataset["url"]
        provider = self.get_dataset_provider(path)
        return provider.read_dataset(get_fs(path), path, keys=keys, dataset=dataset)

    def get_result(self, dataset, result_id):
        path = dataset["url"]
        provider = self.get_dataset_provider(path)
        return provider.get_result(get_fs(path), path, dataset=dataset, result_id=result_id)



# -*- coding: utf-8 -*-

import numpy as np


class Population:
    """A Population is a measure over the cells at given timepoint.

    Parameters
    ----------
    time : int or float
        The time at which the cells where measured.
    p : 1-D array-like
        Measure over the cells at the given timepoint.
    name : str
        Optional population name.
    """

    def __init__(self, time, p, name=None):
        self.time = time
        self.p = np.asarray(p, dtype=np.float64)
        self.name = name

    def normalize(self):
        """Make the measure sum to 1, i.e. be a probability distribution over cells."""
        self.p = self.p / self.p.sum()

    def make_binary(self):
        """Set non-zero values to 1."""
        p = np.zeros(len(self.p))
        p[self.p > 0.0] = 1.0
        self.p = p

    @staticmethod
    def get_missing_population(*populations):
        initial_p_sum = np.array([pop.p for pop in populations]).T.sum(axis=1)
        missing_cells = np.where(initial_p_sum == 0)[0]
        if len(missing_cells) > 0:
            missing_cells_p = np.zeros_like(populations[0].p)
            missing_cells_p[missing_cells] = 1.0
            return Population(populations[0].time, missing_cells_p, "Other")
        return None

    @staticmethod
    def copy(*populations, normalize=None, add_missing=False):
        populations_copy = []
        for pop in populations:
            pop_copy = Population(pop.time, pop.p, pop.name)
            populations_copy.append(pop_copy)
        populations = populations_copy
        if add_missing:
            # add "other" population if any cells are missing across all populations
            missing_pop = Population.get_missing_population(*populations)
            if missing_pop is not None:
                populations.append(missing_pop)
        if normalize or normalize is False:
            for pop in populations:
                if normalize:
                    pop.normalize()
                else:
                    pop.make_binary()
        return populations



import os
import datetime

from bson import ObjectId
from pymongo import MongoClient

from cirrocumulus.abstract_db import AbstractDB
from cirrocumulus.util import dumps, get_email_domain, get_fs

from .envir import (
    CIRRO_AUTH_CLIENT_ID,
    CIRRO_DB_URI,
    CIRRO_JOB_RESULTS,
    CIRRO_TEST,
    SERVER_CAPABILITY_ADD_DATASET,
    SERVER_CAPABILITY_DELETE_DATASET,
    SERVER_CAPABILITY_EDIT_DATASET,
    SERVER_CAPABILITY_FEATURE_SETS,
    SERVER_CAPABILITY_JOBS,
    SERVER_CAPABILITY_LINKS,
    SERVER_CAPABILITY_RENAME_CATEGORIES,
)
from .invalid_usage import InvalidUsage
from .job_api import save_job_result_to_file


def format_doc(d):
    d["id"] = str(d.pop("_id"))
    return d


class MongoDb(AbstractDB):
    def __init__(self):
        super().__init__()
        if bool(os.environ.get(CIRRO_TEST)):
            import mongomock

            self.client = mongomock.MongoClient(os.environ[CIRRO_DB_URI])
        else:
            self.client = MongoClient(os.environ[CIRRO_DB_URI])
        self.db = self.client.get_default_database()
        self.db.categories.create_index("dataset_id")
        self.db.views.create_index("dataset_id")
        self.db.feature_sets.create_index("dataset_id")
        self.db.jobs.create_index("dataset_id")
        self.db.datasets.create_index("readers")
        self.fs = None

    def category_names(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.categories
        results = {}
        for doc in collection.find(dict(dataset_id=dataset_id)):
            category = doc["category"]
            cat_results = results.get(category)
            if cat_results is None:
                cat_results = {}
                results[category] = cat_results
            cat_results[doc["original"]] = {
                "newValue": doc.get("new"),
                "negativeMarkers": doc.get("negativeMarkers"),
                "positiveMarkers": doc.get("positiveMarkers"),
                "color": doc.get("color"),
            }
        return results

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        if not self.capabilities()[SERVER_CAPABILITY_RENAME_CATEGORIES]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.categories
        key = str(dataset_id) + "-" + str(category) + "-" + str(original_value)
        update["category"] = category
        update["dataset_id"] = dataset_id
        update["original"] = original_value
        update["last_updated"] = datetime.datetime.utcnow()
        if "newValue" in update:  # backwards compatibility
            update["new"] = update.pop("newValue")
        collection.update_one(dict(cat_id=key), {"$set": update}, upsert=True)

    def user(self, email):
        collection = self.db.users
        doc = collection.find_one(dict(email=email))
        if doc is None:
            collection.insert_one(dict(email=email))
            return {"id": email}
        else:
            return {"id": email, "importer": doc.get("importer", False)}

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        collection = self.db.datasets
        auth_client_id = os.environ.get(CIRRO_AUTH_CLIENT_ID)

        if auth_client_id is None:  # allow unregistered URL
            try:
                dataset_id.index("://")
                return {"id": dataset_id, "name": dataset_id, "url": dataset_id}
            except ValueError:
                pass
        doc = collection.find_one(dict(_id=ObjectId(dataset_id)))
        if doc is None:
            raise InvalidUsage("Please provide a valid id", 400)
        readers = doc.get("readers")
        domain = get_email_domain(email)
        if email not in readers and domain not in readers and "" not in readers:
            raise InvalidUsage("Not authorized", 403)
        if ensure_owner and email not in doc["owners"]:
            raise InvalidUsage("Not authorized", 403)
        doc["owner"] = "owners" in doc and email in doc.pop("owners")
        doc["id"] = str(doc.pop("_id"))
        return doc

    def datasets(self, email):
        collection = self.db.datasets
        results = []
        domain = get_email_domain(email)
        if domain is None:
            query = dict(readers={"$in": [email, ""]})
        else:
            query = dict(readers={"$in": [email, domain]})

        for doc in collection.find(query):
            doc["owner"] = "owners" in doc and email in doc.pop("owners")
            doc["id"] = str(doc.pop("_id"))
            results.append(doc)
        return results

    # views
    def dataset_views(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.views
        results = []
        for doc in collection.find(dict(dataset_id=dataset_id), dict(value=0)):
            results.append(format_doc(doc))
        return results

    def delete_dataset_view(self, email, view_id):
        if not self.capabilities()[SERVER_CAPABILITY_LINKS]:
            return
        collection = self.db.views
        doc = collection.find_one(dict(_id=ObjectId(view_id)))
        self.get_dataset(email, doc["dataset_id"])
        collection.delete_one(dict(_id=ObjectId(view_id)))

    def get_dataset_view(self, email, view_id):
        collection = self.db.views
        doc = collection.find_one(dict(_id=ObjectId(view_id)))
        self.get_dataset(email, doc["dataset_id"])
        return format_doc(doc)

    def upsert_dataset_view(self, email, dataset_id, view):
        if not self.capabilities()[SERVER_CAPABILITY_LINKS]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.views
        view["last_updated"] = datetime.datetime.utcnow()

        if email is not None:
            view["email"] = email
        if dataset_id is not None:
            view["dataset_id"] = dataset_id
        if view.get("id") is None:
            return dict(
                id=str(collection.insert_one(view).inserted_id), last_updated=view["last_updated"]
            )
        else:
            view_id = view.pop("id")
            collection.update_one(dict(_id=ObjectId(view_id)), {"$set": view})
            return dict(id=view_id, last_updated=view["last_updated"])

    def is_importer(self, email):
        # TODO check if user can modify dataset
        user = self.db.users.find_one(dict(email=email))
        if "importer" not in user:
            raise False
        return user["importer"]

    def delete_dataset(self, email, dataset_id):
        if not self.capabilities()[SERVER_CAPABILITY_DELETE_DATASET]:
            return

        self.get_dataset(email, dataset_id, True)
        self.db.datasets.delete_one(dict(_id=ObjectId(dataset_id)))
        self.db.filters.delete_many(dict(dataset_id=dataset_id))
        self.db.views.delete_many(dict(dataset_id=dataset_id))
        self.db.categories.delete_many(dict(dataset_id=dataset_id))
        self.db.feature_sets.delete_many(dict(dataset_id=dataset_id))

        for doc in self.db.jobs.find(dict(dataset_id=dataset_id), dict(result=1)):
            result = doc.get("result")
            self.delete_job_result(result)
        self.db.jobs.delete_many(dict(dataset_id=dataset_id))

    def upsert_dataset(self, email, readers, dataset):
        if dataset.get("id") is None and not self.capabilities()[SERVER_CAPABILITY_ADD_DATASET]:
            return
        if (
            dataset.get("id") is not None
            and not self.capabilities()[SERVER_CAPABILITY_EDIT_DATASET]
        ):
            return
        collection = self.db.datasets
        if readers is not None:
            readers = set(readers)
            if email in readers:
                readers.remove(email)
            readers.add(email)
            dataset["readers"] = list(readers)
        dataset["last_updated"] = datetime.datetime.utcnow()
        if dataset.get("id") is None:  # new dataset
            if email != "":
                user = self.db.users.find_one(dict(email=email))
                if "importer" not in user or not user["importer"]:
                    raise InvalidUsage("Not authorized", 403)
            dataset["owners"] = [email]
            if "readers" not in dataset:
                dataset["readers"] = [email]
            return str(collection.insert_one(dataset).inserted_id)
        else:  # update
            self.get_dataset(email, dataset["id"], True)
            dataset_id = dataset.pop("id")
            collection.update_one(dict(_id=ObjectId(dataset_id)), {"$set": dataset})
            return dataset_id

    def get_feature_sets(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.feature_sets
        results = []
        for doc in collection.find(dict(dataset_id=dataset_id)):
            results.append(format_doc(doc))
        return results

    def delete_feature_set(self, email, dataset_id, set_id):
        if not self.capabilities()[SERVER_CAPABILITY_FEATURE_SETS]:
            return
        collection = self.db.feature_sets
        doc = collection.find_one(dict(_id=ObjectId(set_id)))
        self.get_dataset(email, doc["dataset_id"])
        collection.delete_one(dict(_id=ObjectId(set_id)))

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        if not self.capabilities()[SERVER_CAPABILITY_FEATURE_SETS]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.feature_sets
        entity_update = {}
        entity_update["last_updated"] = datetime.datetime.utcnow()
        if name is not None:
            entity_update["name"] = name
        if features is not None:
            entity_update["features"] = features
        if email is not None:
            entity_update["email"] = email
        if dataset_id is not None:
            entity_update["dataset_id"] = dataset_id
        if category is not None:
            entity_update["category"] = category
        if set_id is None:
            return str(collection.insert_one(entity_update).inserted_id)
        else:
            collection.update_one(dict(_id=ObjectId(set_id)), {"$set": entity_update})
            return set_id

    def get_gridfs(self):
        import gridfs

        if self.fs is None:
            self.fs = gridfs.GridFS(self.db)
        return self.fs

    def create_job(self, email, dataset_id, job_name, job_type, params):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        self.get_dataset(email, dataset_id)
        collection = self.db.jobs
        job_id = str(
            collection.insert_one(
                dict(
                    dataset_id=dataset_id,
                    status="pending",
                    name=job_name,
                    email=email,
                    type=job_type,
                    params=params,
                    submitted=datetime.datetime.utcnow(),
                )
            ).inserted_id
        )
        return job_id

    def get_job(self, email, job_id, return_type):
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)))
        if doc is None:
            return None
        self.get_dataset(email, doc["dataset_id"])
        if return_type == "result":
            result = doc.get("result")
            if isinstance(result, dict) and result.get("url") is not None:
                return result
            else:
                fs = self.get_gridfs()
                return fs.get(ObjectId(doc["result"])).read()
        elif return_type == "status":
            return dict(status=doc["status"])
        elif return_type == "params":
            return dict(params=doc.get("params"))

    def get_jobs(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.jobs
        results = []
        for doc in collection.find(
            dict(dataset_id=dataset_id),
            dict(name=1, status=1, email=1, type=1, submitted=1, readonly=1),
        ):
            results.append(
                dict(
                    id=str(doc["_id"]),
                    name=doc["name"],
                    status=doc["status"],
                    type=doc["type"],
                    email=doc["email"],
                    submitted=doc.get("submitted"),
                    readonly=doc.get("readonly", False),
                )
            )
        return results

    def annotate_job(self, email, job_id, annotations):
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)))
        self.get_dataset(email, doc["dataset_id"])
        collection.update_one(
            dict(_id=ObjectId(job_id)),
            {"$set": dict(annotations=annotations, last_updated=datetime.datetime.utcnow())},
        )

    def update_job(self, email, job_id, status, result):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)))
        self.get_dataset(email, doc["dataset_id"])
        if doc.get("readonly", False):
            raise InvalidUsage("Not authorized", 403)
        if result is not None:
            if os.environ.get(CIRRO_JOB_RESULTS) is not None:  # save to directory
                result = save_job_result_to_file(result, job_id)
            else:
                result = dumps(result, double_precision=2, orient="values")
                result = str(self.get_gridfs().put(result, encoding="ascii"))

        collection.update_one(
            dict(_id=ObjectId(job_id)), {"$set": dict(status=status, result=result)}
        )

    def delete_job_result(self, result):
        if result is not None:
            if isinstance(result, dict) and result.get("url") is not None:
                fs = get_fs(result["url"])
                if fs.exists(result["url"]):
                    fs.rm(result["url"], recursive=True)
            else:
                self.get_gridfs().delete(ObjectId(result))

    def delete_job(self, email, job_id):
        if not self.capabilities()[SERVER_CAPABILITY_JOBS]:
            return
        collection = self.db.jobs
        doc = collection.find_one(dict(_id=ObjectId(job_id)), dict(email=1))
        if doc["email"] == email and not doc.get("readonly", False):
            collection.delete_one(dict(_id=ObjectId(job_id)))
            result = doc.get("result")
            self.delete_job_result(result)
        else:
            raise InvalidUsage("Not authorized", 403)



import os

import requests
import cachecontrol
import google.auth.transport.requests
from flask import request
from google.oauth2 import id_token

from cirrocumulus.auth_exception import AuthException
from cirrocumulus.envir import CIRRO_AUTH_CLIENT_ID


class GoogleAuth:
    def __init__(self):
        session = requests.session()
        self.client_id = os.environ.get(CIRRO_AUTH_CLIENT_ID)
        cached_session = cachecontrol.CacheControl(session)
        self.cached_request = google.auth.transport.requests.Request(session=cached_session)

    def auth(self):
        token = request.headers.get("Authorization")
        if token is not None:
            if not token.startswith("Bearer "):
                raise AuthException("Token should start with Bearer ")
            token = token.split("Bearer ")[1]
        else:
            token = request.args.get("access_token")
        idinfo = id_token.verify_oauth2_token(token, self.cached_request, self.client_id)
        if idinfo["aud"] != self.client_id:
            raise AuthException("Wrong aud")
        if idinfo["iss"] not in ["accounts.google.com", "https://accounts.google.com"]:
            raise AuthException()("Wrong issuer.")
        return idinfo



from collections.abc import Mapping
from enum import Enum
from functools import _find_impl, singledispatch
from types import MappingProxyType
from typing import Callable, Type, TypeVar
from warnings import warn

import zarr
import numpy as np
import pandas as pd
import numcodecs
from packaging import version
from pandas import CategoricalDtype
from scipy import sparse


class WriteWarning(UserWarning):
    pass


T = TypeVar("T")


def _to_fixed_length_strings(value: np.ndarray) -> np.ndarray:
    """Convert variable length strings to fixed length.

    Currently a workaround for https://github.com/zarr-developers/zarr-python/pull/422
    """
    new_dtype = []
    for dt_name, (dt_type, dt_offset) in value.dtype.fields.items():
        if dt_type.kind == "O":
            #  Assuming the objects are str
            size = max(len(x.encode()) for x in value.getfield("O", dt_offset))
            new_dtype.append((dt_name, ("U", size)))
        else:
            new_dtype.append((dt_name, dt_type))
    return value.astype(new_dtype)


def check_key(key):
    """Checks that passed value is a valid h5py key.

    Should convert it if there is an obvious conversion path, error otherwise.
    """
    typ = type(key)
    if issubclass(typ, str):
        return str(key)
    # TODO: Should I try to decode bytes? It's what h5py would do,
    # but it will be read out as a str.
    # elif issubclass(typ, bytes):
    # return key
    else:
        raise TypeError(f"{key} of type {typ} is an invalid key. Should be str.")


# -------------------------------------------------------------------------------
# Generic functions
# -------------------------------------------------------------------------------


@singledispatch
def write_attribute(*args, **kwargs):
    raise NotImplementedError("Unrecognized argument types for `write_attribute`.")


class EncodingVersions(Enum):
    raw = "0.1.0"
    csr_matrix = csc_matrix = "0.1.0"
    dataframe = "0.1.0"

    def check(self, key: str, encoded_version: str):
        if version.parse(encoded_version) > version.parse(self.value):
            warn(
                f"The supported version for decoding {self.name} is {self.value}, "
                f"but a {self.name} with version {encoded_version} "
                f"was encountered at {key}.",
                FutureWarning,
            )


def _write_method(cls: Type[T]) -> Callable[[zarr.Group, str, T], None]:
    return _find_impl(cls, ZARR_WRITE_REGISTRY)


@write_attribute.register(zarr.Group)
def write_attribute_zarr(f, key, value, dataset_kwargs=MappingProxyType({})):
    if key in f:
        del f[key]
    _write_method(type(value))(f, key, value, dataset_kwargs)


def write_mapping(f, key, value: Mapping, dataset_kwargs=MappingProxyType({})):
    for sub_k, sub_v in value.items():
        if not isinstance(key, str):
            warn(
                f"dict key {key} transformed to str upon writing to zarr, using "
                "string keys is recommended.",
                WriteWarning,
            )
        write_attribute(f, f"{key}/{sub_k}", sub_v, dataset_kwargs)


def write_dataframe(z, key, df, dataset_kwargs=MappingProxyType({})):
    # Check arguments
    for reserved in ("__categories", "_index"):
        if reserved in df.columns:
            raise ValueError(f"{reserved!r} is a reserved name for dataframe columns.")

    col_names = [check_key(c) for c in df.columns]

    if df.index.name is not None:
        index_name = df.index.name
    else:
        index_name = "_index"
    index_name = check_key(index_name)

    group = z.create_group(key)
    group.attrs["encoding-type"] = "dataframe"
    group.attrs["encoding-version"] = EncodingVersions.dataframe.value
    group.attrs["column-order"] = col_names
    group.attrs["_index"] = index_name

    write_series(group, index_name, df.index, dataset_kwargs)
    for col_name, (_, series) in zip(col_names, df.items()):
        write_series(group, col_name, series, dataset_kwargs)


def write_series(group, key, series, dataset_kwargs=MappingProxyType({})):
    if series.dtype == object:
        group.create_dataset(
            key,
            shape=series.shape,
            dtype=object,
            object_codec=numcodecs.VLenUTF8(),
            **dataset_kwargs,
        )
        group[key][:] = series.values
    elif isinstance(series.dtype, CategoricalDtype):
        # This should work for categorical Index and Series
        categorical: pd.Categorical = series.values
        categories: np.ndarray = categorical.categories.values
        codes: np.ndarray = categorical.codes
        category_key = f"__categories/{key}"

        write_array(group, category_key, categories, dataset_kwargs=dataset_kwargs)
        write_array(group, key, codes, dataset_kwargs=dataset_kwargs)

        group[key].attrs["categories"] = category_key
        # Must coerce np.bool_ to bool for json writing
        group[category_key].attrs["ordered"] = bool(categorical.ordered)
    else:
        write_array(group, key, series.values, dataset_kwargs=dataset_kwargs)


def write_not_implemented(f, key, value, dataset_kwargs=MappingProxyType({})):
    # If it’s not an array, try and make it an array. If that fails, pickle it.
    # Maybe rethink that, maybe this should just pickle,
    # and have explicit implementations for everything else
    raise NotImplementedError(
        f"Failed to write value for {key}, since a writer for type {type(value)}"
        f" has not been implemented yet."
    )


def write_list(g, key, value, dataset_kwargs=MappingProxyType({})):
    write_array(g, key, np.array(value), dataset_kwargs)


def write_array(g, key, value, dataset_kwargs=MappingProxyType({})):
    if value.dtype == object:
        g.create_dataset(
            key,
            shape=value.shape,
            dtype=object,
            object_codec=numcodecs.VLenUTF8(),
            **dataset_kwargs,
        )
        g[key][:] = value
    elif value.dtype.kind == "V":
        # Structured dtype
        g.create_dataset(key, data=_to_fixed_length_strings(value), **dataset_kwargs)
    else:
        g.create_dataset(key, data=value, **dataset_kwargs)


# TODO: Not working quite right


def write_scalar(f, key, value, dataset_kwargs=MappingProxyType({})):
    f.create_dataset(key, data=np.array(value), **dataset_kwargs)


def write_none(f, key, value, dataset_kwargs=MappingProxyType({})):
    pass


# TODO: Figure out what to do with dataset_kwargs for these


def write_csr(f, key, value: sparse.csr_matrix, dataset_kwargs=MappingProxyType({})):
    group = f.create_group(key)
    group.attrs["encoding-type"] = "csr_matrix"
    group.attrs["encoding-version"] = EncodingVersions.csr_matrix.value
    group.attrs["shape"] = value.shape
    write_array(group, "data", value.data, dataset_kwargs=dataset_kwargs)
    write_array(group, "indices", value.indices, dataset_kwargs=dataset_kwargs)
    write_array(group, "indptr", value.indptr, dataset_kwargs=dataset_kwargs)


def write_csc(f, key, value: sparse.csc_matrix, dataset_kwargs=MappingProxyType({})):
    group = f.create_group(key)
    group.attrs["encoding-type"] = "csc_matrix"
    group.attrs["encoding-version"] = EncodingVersions.csc_matrix.value
    group.attrs["shape"] = value.shape
    write_array(group, "data", value.data, dataset_kwargs=dataset_kwargs)
    write_array(group, "indices", value.indices, dataset_kwargs=dataset_kwargs)
    write_array(group, "indptr", value.indptr, dataset_kwargs=dataset_kwargs)


def write_raw(f, key, value, dataset_kwargs=MappingProxyType({})):
    group = f.create_group(key)
    group.attrs["encoding-type"] = "raw"
    group.attrs["encoding-version"] = EncodingVersions.raw.value
    group.attrs["shape"] = value.shape
    write_attribute(group, "X", value.X, dataset_kwargs)
    write_attribute(group, "var", value.var, dataset_kwargs)
    write_attribute(group, "varm", value.varm, dataset_kwargs)


ZARR_WRITE_REGISTRY = {
    type(None): write_none,
    Mapping: write_mapping,
    object: write_not_implemented,
    np.ndarray: write_array,  # Possibly merge with write_series
    list: write_list,
    pd.DataFrame: write_dataframe,
    # Raw: write_raw,
    # object: write_not_implemented,
    # h5py.Dataset: write_basic,
    # type(None): write_none,
    str: write_scalar,
    float: write_scalar,
    np.floating: write_scalar,
    bool: write_scalar,
    np.bool_: write_scalar,
    int: write_scalar,
    np.integer: write_scalar,
    sparse.csr_matrix: write_csr,
    sparse.csc_matrix: write_csc,
}



import os
import argparse

import anndata

import cirrocumulus
from cirrocumulus.anndata_dataset import AnndataDataset
from cirrocumulus.envir import (
    CIRRO_AUTH,
    CIRRO_CELL_ONTOLOGY,
    CIRRO_COMPRESS,
    CIRRO_DATABASE,
    CIRRO_JOB_RESULTS,
    CIRRO_JOB_TYPE,
)
from cirrocumulus.io_util import SPATIAL_HELP, add_spatial, filter_markers, get_markers
from cirrocumulus.local_db_api import LocalDbAPI
from cirrocumulus.util import get_fs


def configure_app(app, list_of_dataset_paths, spatial_directories, marker_paths):
    from cirrocumulus.api import dataset_api
    from cirrocumulus.no_auth import NoAuth

    try:
        from cirrocumulus.parquet_dataset import ParquetDataset

        dataset_api.add(ParquetDataset())
    except ModuleNotFoundError:
        pass
    try:
        from cirrocumulus.zarr_dataset import ZarrDataset

        dataset_api.add(ZarrDataset())
    except ModuleNotFoundError:
        pass
    app.config[CIRRO_AUTH] = NoAuth()
    os.environ[CIRRO_JOB_TYPE + "de"] = "cirrocumulus.job_api.run_de"
    os.environ[CIRRO_JOB_TYPE + "ot_trajectory"] = "cirrocumulus.job_api.run_ot_trajectory"
    anndata_dataset = AnndataDataset()
    dataset_api.add(anndata_dataset)
    dataset_ids = []
    for dataset_paths in list_of_dataset_paths:
        dataset_paths = dataset_paths.split(",")
        dataset_id = dataset_paths[0]
        dataset_ids.append(dataset_id)
        if len(dataset_paths) > 1:
            datasets = []
            for i in range(len(dataset_paths)):
                dataset = anndata_dataset.get_data(get_fs(dataset_paths[i]), dataset_paths[i])
                if "group" not in dataset.var:
                    dataset.var["group"] = dataset.uns.get("name", "dataset {}".format(i + 1))
                datasets.append(dataset)
            adata = anndata.concat(datasets, axis=1, label="group", merge="unique")
            dataset.obsm = datasets[0].obsm
            adata.var.index = adata.var.index.str.replace("/", "_")
            adata.var_names_make_unique()
            anndata_dataset.add_data(dataset_id, adata)

    app.config[CIRRO_DATABASE] = LocalDbAPI(dataset_ids)

    if spatial_directories is not None and len(spatial_directories) > 0:
        for i in range(len(spatial_directories)):
            spatial_directory = spatial_directories[i]
            if spatial_directory != "":
                adata = anndata_dataset.get_data(get_fs(dataset_ids[i]), dataset_ids[i])
                if not add_spatial(adata, spatial_directory):
                    print("No spatial data found in {}".format(spatial_directory))

    if marker_paths is not None and len(marker_paths) > 0:
        markers = get_markers(marker_paths)
        for dataset_id in dataset_ids:
            d = anndata_dataset.get_data(get_fs(dataset_id), dataset_id)
            existing_markers = d.uns.get("markers", [])
            markers += existing_markers
            # remove genes in dict that are not in dataset
            d.uns["markers"] = filter_markers(d, markers)


def create_app():
    from flask import Flask, send_from_directory
    from flask_compress import Compress

    from cirrocumulus.api import cirro_blueprint

    app = Flask(
        __name__, static_folder=os.path.join(cirrocumulus.__path__[0], "client"), static_url_path=""
    )
    app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0
    app.register_blueprint(cirro_blueprint, url_prefix="/api")

    @app.route("/")
    def root():
        return send_from_directory(
            os.path.abspath(os.path.join(app.root_path, "client")), "index.html"
        )

    if bool(os.environ.get(CIRRO_COMPRESS, "true")):
        Compress(app)
    return app


def create_parser(description=False):
    parser = argparse.ArgumentParser(description="Run cirrocumulus" if description else None)
    parser.add_argument(
        "dataset",
        help="Path(s) to dataset in h5ad, loom, Seurat, TileDB, zarr, or STAR-Fusion format. Separate multiple datasets with "
        "a comma instead of a space in order to join datasets by cell id",
        nargs="+",
    )
    parser.add_argument("--spatial", help=SPATIAL_HELP, nargs="*")
    parser.add_argument(
        "--markers",
        help='Path(s) to JSON file that maps name to features. For example {"a":["gene1", "gene2"], "b":["gene3"]}',
        nargs="*",
    )
    parser.add_argument(
        "--host", help="Host IP address"
    )  # set to 0.0.0.0 to make it accessible from other computers WITHOUT SECURITY.

    parser.add_argument("--port", help="Server port", default=5000, type=int)
    parser.add_argument(
        "--no-open", dest="no_open", help="Do not open your web browser", action="store_true"
    )
    parser.add_argument(
        "--results", help="URL to save user computed results (e.g. differential expression)"
    )
    parser.add_argument("--ontology", help="Path to ontology in OBO format for annotation")
    parser.add_argument(
        "--tmap", help="Path(s) to transport maps directory computed with WOT", nargs="*"
    )
    return parser


def main(argsv):
    args = create_parser(True).parse_args(argsv)
    if args.results is not None:
        os.environ[CIRRO_JOB_RESULTS] = args.results
    else:
        os.environ[CIRRO_JOB_RESULTS] = os.path.join(
            os.path.dirname(args.dataset[0].rstrip("/")), "results"
        )
    get_fs(os.environ[CIRRO_JOB_RESULTS]).makedirs(os.environ[CIRRO_JOB_RESULTS], exist_ok=True)
    if args.ontology is not None:
        os.environ[CIRRO_CELL_ONTOLOGY] = args.ontology
    app = create_app()
    configure_app(app, args.dataset, args.spatial, args.markers)
    if args.tmap is not None:
        tmaps = []
        unique_names = set()
        for tmap_dir in args.tmap:
            name = os.path.basename(tmap_dir.rstrip("/"))
            counter = 1
            while name in unique_names:
                name = f"{name}-{counter}"
                counter = counter + 1
            tmaps.append({"path": tmap_dir, "name": name})
        app.config[CIRRO_DATABASE].add_metadata(None, {"ot": {"tmaps": tmaps}})
    if not args.no_open:
        import webbrowser

        import requests

        host = args.host if args.host is not None else "http://127.0.0.1"
        url = host + ":" + str(args.port)
        try:
            if requests.get(url).ok:
                import sys

                sys.exit("Address already in use")
        except:
            pass
        webbrowser.open(url)
    from flask import cli

    cli.show_server_banner = lambda *_: None  # suppress warning message
    app.run(host=args.host, port=args.port, debug=False)


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



import os
import json

import pandas as pd
import tiledb
import scipy.sparse
from anndata import AnnData

from cirrocumulus.abstract_dataset import AbstractDataset


class TileDBDataset(AbstractDataset):
    def get_suffixes(self):
        return ["cxg"]

    def get_schema(self, filesystem, path):
        # path is to directory
        schema_dict = {"version": "1.0.0"}
        schema_dict["markers"] = []

        with tiledb.Array(os.path.join(path, "X"), mode="r") as array:
            schema_dict["shape"] = array.shape
        annotations = {}
        for ax in ["obs"]:
            with tiledb.open(os.path.join(path, ax), mode="r") as array:
                schema_hints = (
                    json.loads(array.meta["cxg_schema"]) if "cxg_schema" in array.meta else {}
                )
                if type(schema_hints) is not dict:
                    raise TypeError("Array schema was malformed.")
                cols = []
                for attr in array.schema:
                    schema = dict(name=attr.name, writable=False)
                    type_hint = schema_hints.get(attr.name, {})
                    # type hints take precedence
                    if "type" in type_hint:
                        schema["type"] = type_hint["type"]
                        if schema["type"] == "categorical" and "categories" in type_hint:
                            schema["categories"] = type_hint["categories"]
                    # else:
                    #     schema.update(get_schema_type_hint_from_dtype(attr.dtype))
                    cols.append(schema)

                annotations[ax] = dict(columns=cols)

                if "index" in schema_hints:
                    annotations[ax].update({"index": schema_hints["index"]})
        obs = []
        obs_cat = []
        category_order = {}
        for c in annotations["obs"]["columns"]:
            if c["name"] == "name_0":  # index
                continue
            if "type" in c and c["type"] == "categorical":
                obs_cat.append(c["name"])
                category_order[c["name"]] = c["categories"]
            else:
                obs.append(c["name"])
        schema_dict["obsIndex"] = annotations["obs"].get("index", "name_0")
        schema_dict["obs"] = obs
        schema_dict["obsCat"] = obs_cat
        schema_dict["categoryOrder"] = category_order

        # with tiledb.Array(os.path.join(path, 'cxg_group_metadata'), mode="r") as gmd:
        #     # cxg_version = gmd.meta["cxg_version"]
        #     # # version 0.1 used a malformed/shorthand semver string.
        #     # if cxg_version == "0.1" or cxg_version == "0.2.0":
        #     #     cxg_properties = json.loads(gmd.meta["cxg_properties"])
        #     colors = json.loads(gmd.meta["cxg_category_colors"]) if "cxg_category_colors" in gmd.meta else dict()
        embeddings_path_type = []
        tiledb.ls(
            os.path.join(path, "emb"), lambda path, type: embeddings_path_type.append((path, type))
        )
        embeddings = []
        schema_dict["embeddings"] = embeddings
        for path_type in embeddings_path_type:
            if path_type[1] == "array":
                with tiledb.open(path_type[0], mode="r") as array:
                    name = os.path.basename(path_type[0])
                    dimensions = array.ndim
                    if dimensions > 2:
                        embeddings.append({"name": name, "dimensions": 3})
                        embeddings.append({"name": name, "dimensions": 2})
                    elif dimensions == 2:
                        embeddings.append({"name": name, "dimensions": 2})

        with tiledb.open(os.path.join(path, "var"), mode="r") as array:
            schema_dict["var"] = pd.Index(array.query(attrs=["name_0"])[:]["name_0"])
        return schema_dict

    def read_dataset(self, filesystem, path, keys=None, dataset=None):
        keys = keys.copy()
        var_keys = keys.pop("X", [])
        obs_keys = keys.pop("obs", [])
        basis_keys = keys.pop("basis", [])
        dataset_info = self.get_dataset_info(filesystem, path)
        X = None
        obs = None
        var = None
        obsm = {}
        if len(var_keys) > 0:
            with tiledb.open(os.path.join(path, "X"), mode="r") as array:
                if array.schema.sparse:
                    raise ValueError("Sparse data not supported")
                var_names = dataset_info["var"]
                indices = var_names.get_indexer_for(var_keys).tolist()
                X = array.multi_index[:, indices][""]
                X = scipy.sparse.csc_matrix(X)
                var = pd.DataFrame(index=var_keys)
        if len(obs_keys) > 0:
            obs = pd.DataFrame()
            _obs_keys = []

            for key in obs_keys:
                if key == "index":
                    _obs_keys.append(dataset_info.get("obsIndex", "name_0"))
                else:
                    _obs_keys.append(key)
            with tiledb.open(os.path.join(path, "obs"), mode="r") as array:
                ordered_dict = array.query(attrs=_obs_keys)[:]
            for key in ordered_dict:
                obs[key] = ordered_dict[key]

        if len(basis_keys) > 0:
            for key in basis_keys:
                with tiledb.open(os.path.join(path, "emb", key), mode="r") as array:
                    obsm[key] = array[:]
                    if X is None:
                        X = scipy.sparse.coo_matrix(([], ([], [])), shape=(array.shape[0], 0))
        return AnnData(X=X, obs=obs, var=var, obsm=obsm)



# Abstract database class supporting datasets, categories, views, feature sets, users, jobs, and job results
# Only methods server and datasets must be supported in 'client' mode
import os

from cirrocumulus.envir import (
    SERVER_CAPABILITY_ADD_DATASET,
    SERVER_CAPABILITY_DELETE_DATASET,
    SERVER_CAPABILITY_EDIT_DATASET,
    SERVER_CAPABILITY_FEATURE_SETS,
    SERVER_CAPABILITY_JOBS,
    SERVER_CAPABILITY_LINKS,
    SERVER_CAPABILITY_RENAME_CATEGORIES,
)


def to_bool(s):
    return s.lower() in ["true", "1"]


class AbstractDB:
    def capabilities(self):  # allow everything
        c = {}
        c[SERVER_CAPABILITY_RENAME_CATEGORIES] = to_bool(
            os.environ.get(SERVER_CAPABILITY_RENAME_CATEGORIES, "True")
        )
        c[SERVER_CAPABILITY_JOBS] = to_bool(os.environ.get(SERVER_CAPABILITY_JOBS, "True"))
        c[SERVER_CAPABILITY_FEATURE_SETS] = to_bool(
            os.environ.get(SERVER_CAPABILITY_FEATURE_SETS, "True")
        )
        c[SERVER_CAPABILITY_LINKS] = to_bool(os.environ.get(SERVER_CAPABILITY_LINKS, "True"))
        c[SERVER_CAPABILITY_EDIT_DATASET] = to_bool(
            os.environ.get(SERVER_CAPABILITY_EDIT_DATASET, "True")
        )
        c[SERVER_CAPABILITY_ADD_DATASET] = to_bool(
            os.environ.get(SERVER_CAPABILITY_ADD_DATASET, "True")
        )
        c[SERVER_CAPABILITY_DELETE_DATASET] = to_bool(
            os.environ.get(SERVER_CAPABILITY_DELETE_DATASET, "True")
        )
        return c

    def datasets(self, email):
        """Gets list of available datasets.

        Args:
            email: User email or None

        Returns:
            A list of dicts. Example:
            [{"id": "dataset_id",
             "name": "dataset_name"
             "title": "dataset_title",
             "owner": false,
             "url": "gs://xx/my_dataset".
             "species": "Mus Musculus"
            }]
        """
        raise NotImplementedError()

    def category_names(self, email, dataset_id):
        """Gets a dict of renamed categories for a dataset.

        Args:
              email: User email or None
              dataset_id: Dataset id

        Returns:
            A dict that maps category->category_value->{}. Example:

            {"louvain":{"1":{"color":"red"}}
        """
        raise NotImplementedError()

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        """Upserts a category name.

        Args:
             email: User email or None
             category: Category in dataset (e.g. louvain)
             dataset_id: Dataset id
             original_value: Original category value (e.g. "1")
             update: Update object optionally containing color, positiveMarkers, negativeMarkers, newValue
        """
        raise NotImplementedError()

    def user(self, email):
        """Gets metadata about a user.

        Args:
           email: User email

        Returns:
            A dict with the keys id and importer. An importer can add datasets to cirrocumulus. Example:
            {"id":"user_id",
            "importer":false
            }
        """
        raise NotImplementedError()

    # views
    def dataset_views(self, email, dataset_id):
        """Gets list of saved dataset views (saved visualization states)

        Args:
              email: User email or None
              dataset_id: Dataset id

         Returns:
             List of dicts. Example:
             [{"id": "view id",
               "name": "view name"}]
        """
        raise NotImplementedError()

    def delete_dataset_view(self, email, view_id):
        """Delete a saved view.

        Args:
            email: User email or None
            view_id: View id
        """
        raise NotImplementedError()

    def get_dataset_view(self, email, view_id):
        """Gets detailed information for a saved dataset view.

        Args:
            email: User email or None
            view_id: View id

        Returns:
            List of dicts containing id and name. Example:
            [{"id": "view id",
              "name": "view name",
              "value": "JSON encoded state"
              "notes": "view notes"
              "email": "View creator email"
        """
        raise NotImplementedError()

    def upsert_dataset_view(self, email, dataset_id, view):
        """Upserts a dataset view View should have id (for update), name, value, and any other
        additional fields to store.

        Args:
              email: User email or None
              dataset_id: Dataset id
              view: View to upsert

         Returns:
            dict with id, last_updated
        """
        raise NotImplementedError()

    def delete_dataset(self, email, dataset_id):
        """Deletes a dataset.

        Args:
            email: User email or None
            dataset_id: Dataset id
        """
        raise NotImplementedError()

    def upsert_dataset(self, email, readers, dataset):
        """Upserts a dataset. If dataset.id is None then a new dataset is inserted.
        Dataset should have name, url, description, title, species, and any other additional fields to store
        Args:
              email: User email or None
              readers: List of allowed readers
              dataset: Dataset to upsert

         Returns:
            Upserted dataset id
        """
        raise NotImplementedError()

    def get_feature_sets(self, email, dataset_id):
        """Gets saved feature sets.

        Args:
              email: User email or None
              dataset_id: Dataset id

        Returns:
            List of dicts. Example:
            [{"id": "set_id",
              "category": "set category",
              "name": "set name",
              "features: ["gene1", "gene2"]}]
        """
        raise NotImplementedError()

    def delete_feature_set(self, email, dataset_id, set_id):
        """Deletes a saved feature set.

        Args:
              email: User email or None
              dataset_id: Dataset id
              set_id: Feature set id
        """
        raise NotImplementedError()

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        """Upserts a feature set.

        Args:
            email: User email or None
            dataset_id: Dataset id
            set_id: Feature set id
            category: Set category
            name: Set name
            features: List of features

        Returns:
            Upserted id
        """
        raise NotImplementedError()

    def create_job(self, email, dataset_id, job_name, job_type, params):
        """Creates a job.

        Args:
         email: User email or None
         dataset_id: Dataset id
         job_name: Job name
         job_type: Job type
         params: JSON encoded job params

        Returns:
         job id
        """
        raise NotImplementedError()

    def get_job(self, email, job_id, return_type):
        """Gets a job.

        Args:
           email: User email or None
           job_id: Job id
           return_type: One of "result", "status", or "params"

        Returns:
           The job
        """
        raise NotImplementedError()

    def get_jobs(self, email, dataset_id):
        """Gets a list of all jobs for a dataset.

        Args:
            email: User email or None
            dataset_id: Dataset id

        Returns:
            List of jobs
        """
        raise NotImplementedError()

    def delete_job(self, email, job_id):
        """Deletes a job.

        Args:
            email: User email or None
            job_id: Job id
        """
        raise NotImplementedError()

    def update_job(self, email, job_id, status, result):
        """Updates job info.

        Args:
            email: User email or None
            job_id: Job id
            status: Job status
            result: Job result
        """
        raise NotImplementedError()



import zarr

from cirrocumulus.anndata_util import ADATA_MODULE_UNS_KEY, get_pegasus_marker_keys
from cirrocumulus.anndata_zarr import write_attribute
from cirrocumulus.util import dumps


def save_dataset_zarr(dataset, schema, output_directory, filesystem, whitelist):
    module_dataset = None
    if dataset.uns.get(ADATA_MODULE_UNS_KEY) is not None:
        module_dataset = dataset.uns[ADATA_MODULE_UNS_KEY]
        module_dataset.var.index.name = "id"

    dataset.obs.index.name = "id"
    dataset.var.index.name = "id"
    dataset.strings_to_categoricals()
    if module_dataset is not None:
        module_dataset.strings_to_categoricals()

    dataset.uns["cirro-schema"] = dumps(schema, double_precision=2, orient="values")
    group = zarr.open_group(filesystem.get_mapper(output_directory), mode="a")

    if whitelist["x"]:
        write_attribute(group, "X", dataset.X)
        for layer in dataset.layers.keys():
            write_attribute(group, "layers/{}".format(layer), dataset.layers[layer])
        if module_dataset is not None:
            write_attribute(group, "uns/module/X", module_dataset.X)
            write_attribute(group, "uns/module/var", module_dataset.var)
    if whitelist["obs"]:
        write_attribute(group, "obs", dataset.obs)
    if whitelist["obsm"]:
        write_attribute(group, "obsm", dataset.obsm)

    pg_marker_keys = get_pegasus_marker_keys(dataset)
    for key in list(dataset.varm.keys()):
        if key not in pg_marker_keys:
            del dataset.varm[key]
    write_attribute(group, "varm", dataset.varm)
    write_attribute(group, "var", dataset.var)
    # uns_whitelist = set(['module', 'cirro-schema'])
    # keep DE results and colors
    # sc_marker_keys = get_scanpy_marker_keys(dataset)
    # for key in list(dataset.uns.keys()):
    #     if key in uns_whitelist:
    #         continue
    #     keep = False
    #     if key in sc_marker_keys:
    #         keep = True
    #     elif key.endswith('_colors'):
    #         field = key[0:len(key) - len('_colors')]
    #         if field in dataset.obs:
    #             keep = True
    #     if not keep:
    #         del dataset.uns[key]

    for key in list(dataset.uns.keys()):
        # need to write individual groups so don't overwrite uns
        write_attribute(group, "uns/{}".format(key), dataset.uns[key])



import os
import logging

import numpy as np
import pyarrow as pa
import scipy.sparse
import pyarrow.parquet as pq

from cirrocumulus.util import dumps


logger = logging.getLogger("cirro")


def write_pq(d, output_dir, name, filesystem, write_statistics=True, row_group_size=None):
    filesystem.makedirs(output_dir, exist_ok=True)
    pq.write_table(
        pa.Table.from_pydict(d),
        os.path.join(output_dir, name + ".parquet"),
        write_statistics=write_statistics,
        row_group_size=row_group_size,
        filesystem=filesystem,
    )


def save_dataset_pq(dataset, schema, output_directory, filesystem, whitelist):
    X_dir = os.path.join(output_directory, "X")
    obs_dir = os.path.join(output_directory, "obs")
    obsm_dir = os.path.join(output_directory, "obsm")
    filesystem.makedirs(X_dir, exist_ok=True)
    filesystem.makedirs(obs_dir, exist_ok=True)
    filesystem.makedirs(obsm_dir, exist_ok=True)

    with filesystem.open(
        os.path.join(output_directory, "index.json.gz"), "wt", compression="gzip"
    ) as f:
        f.write(dumps(schema, double_precision=2, orient="values"))
        if whitelist["x"]:
            save_adata_X(dataset, X_dir, filesystem, whitelist=whitelist["x_keys"])
            for layer in dataset.layers.keys():
                layer_dir = os.path.join(output_directory, "layers", layer)
                filesystem.makedirs(layer_dir, exist_ok=True)
                save_adata_X(dataset, layer_dir, filesystem, layer, whitelist=whitelist["x_keys"])
        if whitelist["obs"]:
            save_data_obs(dataset, obs_dir, filesystem, whitelist=whitelist["obs_keys"])
        if whitelist["obsm"]:
            save_data_obsm(dataset, obsm_dir, filesystem, whitelist=whitelist["obsm_keys"])


def save_adata_X(adata, X_dir, filesystem, layer=None, whitelist=None):
    adata_X = adata.X if layer is None else adata.layers[layer]
    names = adata.var.index
    is_sparse = scipy.sparse.issparse(adata_X)
    output_dir = X_dir
    for j in range(adata_X.shape[1]):
        filename = names[j]
        if whitelist is None or filename in whitelist:
            X = adata_X[:, j]
            if is_sparse:
                X = X.toarray().flatten()

            if is_sparse:
                indices = np.where(X != 0)[0]
                values = X[indices]
                write_pq(dict(index=indices, value=values), output_dir, filename, filesystem)
            else:
                write_pq(dict(value=X), output_dir, filename, filesystem)
            if j > 0 and (j + 1) % 1000 == 0:
                logger.info("Wrote adata X {}/{}".format(j + 1, adata_X.shape[1]))


def save_data_obsm(adata, obsm_dir, filesystem, whitelist):
    logger.info("writing adata obsm")

    for name in adata.obsm.keys():
        if whitelist is None or name in whitelist:
            m = adata.obsm[name]
            dim = m.shape[1]
            d = {}
            for i in range(dim):
                d[name + "_" + str(i + 1)] = m[:, i].astype("float32")
            write_pq(d, obsm_dir, name, filesystem)


def save_data_obs(adata, obs_dir, filesystem, whitelist=None):
    logger.info("writing adata obs")
    for name in adata.obs:
        if whitelist is None or name in whitelist:
            value = adata.obs[name]
            write_pq(dict(value=value), obs_dir, name, filesystem)
    write_pq(dict(value=adata.obs.index.values), obs_dir, "index", filesystem)
